<?php
Namespace App\Model\Admin;

use Symfony\Component\HttpFoundation\RequestStack;
use Symfony\Component\Routing\RouterInterface;
use Symfony\Component\Translation\TranslatorInterface;
use App\Model\HomegearDevice as BaseDevice;

/**
 * Provides admin functions for Homegear device management
 *
 */
class HomegearDevice extends BaseDevice
{
    // define devices for heating programs
    // unfortunately there is no way around this currently, as we cannot guarantee internet connection
    private $familiesHeatingProgram = [0, 4, 24];
    private $devicesHeatingProgram = [
                                        'single' => ['HM-CC-TC', 'HM-CC-RT-DN', 'BC-RT-TRX-CYG', 'BC-RT-TRX-CYG-2', 'BC-RT-TRX-CYG-3', 'BC-RT-TRX-CYN', 'BC-TC-C-WM-2', 'BC-TC-C-WM-4'], 
                                        'multiple' => ['HM-TC-IT-WM-W-EU','HMIP-ETRV', 'HMIP-ETRV-2', 'HMIP-ETRV-2-UK', 'HMIP-ETRV-B', 'HMIP-ETRV-B1', 'HMIP-ETRV-B-UK', 'HMIP-ETRV-C', 'HMIP-ETRV-C-2', 'HMIP-ETRV-E', 'HMIP-BWTH', 'HMIP-WTH', 'HMIP-WTH-2', 'HMIP-BWTH', 'HMIP-BWTH24', 'HMIP-STH', 'HMIP-STHD', 'HMIPW-WHT', 'HMIPW-WHT-2', 'HMIPW-STH', 'HMIPW-STHD', 'HEATING']
                                     ];
    private $weekdays = array('monday','tuesday','wednesday','thursday','friday','saturday', 'sunday');
    private $translator;
    const componentHeatingProgramTemplates = 'heatingTemplates';
    // list of family IDs which may change the gateway via setInterface()
    const mayChangeGateway = [4];

    public function __construct(RequestStack $requestStack, RouterInterface $router, TranslatorInterface $translator)
    {
        parent::__construct($requestStack, $router);
        $this->translator = $translator;
    }

    /**
     * Load single device
     * 
     * @return array()
     */
    public function load($id = false)
    {
        if ($id === false) { return; }
        try {
            $device = $this->hg->getDeviceDescription(intval($id),-1);
        } catch (\Homegear\HomegearException $e) {
            return;
        }
        $families = $this->_loadFamilyNames();
        $device['FAMILY_NAME'] = $families[$device['FAMILY']];
        // Check for gateway
        $gateway = $this->hg->getDeviceInfo(intval($id),array("INTERFACE"));
        if (array_key_exists('INTERFACE',$gateway)) {
            $alias = $this->hg->getData(self::componentGatewayAlias,$device['FAMILY'].'-'.$gateway['INTERFACE']);
            $device['GATEWAY'] = $gateway['INTERFACE'];
            $device['GATEWAY_ALIAS'] = (isset($alias) ? $alias : '');
        }
        // get the configurable parameters for this family
        $config = $this->getDeviceConfigParams($device['FAMILY'],$device['TYPE_ID']);
        // check if there are configurable parameters for this family
        if (is_array($config) && array_key_exists('CONFIG_PARARMS',$config)) {
            $config = $config['CONFIG_PARARMS'];
            // get the whole paramset for the device
            $paramset = $this->hg->getParamset(intval($id),-1);
            foreach ($config as $k => $v) {
                if (array_key_exists($v['id'],$paramset)) {
                    $config[$k]['value'] = $paramset[$v['id']];
                }
            }
            usort($config, function($a, $b) { return $a['position'] <=> $b['position']; });
            $device['CONFIG'] = $config;
        }
        // TODO: this has to be set by homegear, hardwired temporary!
        $device['ISHUB'] = ($device["FAMILY"] == 18 || $device["FAMILY"] == 8 ? 1 : 0);
        $device['CHANNELS'] = $this->hg->getAllValues(intval($id))[0]['CHANNELS'];
        // get the paramset-descriptions ... needed for rooms/categories
        foreach ($device['CHANNELS'] as $channel => $values) {
            // Get room / categories for channel
            try {
                $associations = $this->hg->getDeviceDescription(intval($id),intval($channel),array("ROOM","CATEGORIES"));
                if (array_key_exists('ROOM',$associations)) {
                    $device['CHANNELS'][$channel]['ROOM'] = $associations['ROOM'];
                }
                if (array_key_exists('CATEGORIES',$associations)) {
                    $device['CHANNELS'][$channel]['CATEGORIES'] = $associations['CATEGORIES'];
                }
                $device['CHANNELS'][$channel]['PARAMSET_DESCRIPTION'] = $this->hg->getParamsetDescription(intval($id),intval($channel),'VALUES');
            } catch(\Homegear\HomegearException $e) {
                $device['CHANNELS'][$channel]['PARAMSET_DESCRIPTION'] = [];
            }
        }
        // check if we have a device with heating programs (only for Homematic/Homematic IP via CCU/MAX!)
        if (in_array($device['FAMILY'],$this->familiesHeatingProgram)) {
            if (in_array(strtoupper($device['TYPE']),$this->devicesHeatingProgram['single']) || in_array(strtoupper($device['TYPE']),$this->devicesHeatingProgram['multiple'])) {
                $device['HEATINGPROGRAMS'] = 1;
            }
        }
        // Check for Homematic IP Access Point heating group device
        if ($device['FAMILY'] == 666 && $device['TYPE'] == 'HEATING') {
            $device['HEATINGPROGRAMS'] = 1;
        }

        $device['SHOW_CHANNELNAMES'] = $this->hg->getMetadata(intval($id), "showChannelNames");
        $device['SHOW_VARIABLENAMES'] = $this->hg->getMetadata(intval($id), "showVariableNames");
        $device['HIDDEN_IN_FRONTEND'] = $this->getDeviceFrontendHidden(intval($id));
        return $device;
    }

    public function loadConfigParams($id = false)
    {
        if ($id === false) { return; }
        // fetch the current config from device ... not implemented for every family, so try & catch
        $this->forceConfigUpdate(intval($id));
        $device = $this->hg->getDeviceDescription(intval($id),-1);
        foreach ($device['CHANNELS'] as $channel) {
            // only get channels with params
            try {
                $params = $this->hg->getParamsetDescription(intval($id),$channel,"MASTER");
                if (isset($params) && count($params) > 0) {
                    foreach ($params as $key => $singleParam) {
                        if ($singleParam['TYPE'] == "STRUCT" || $singleParam['TYPE'] == "ARRAY") {
                            unset($params[$key]);
                        } elseif ($singleParam['FLAGS'] != 1 || ($singleParam['OPERATIONS'] != 2 && $singleParam['OPERATIONS'] != 7)) {
                            // check if param is visible & writeable
                            unset($params[$key]);
                        } else {
                            // fix encoding
                            if (array_key_exists('UNIT', $singleParam)) {
                                $params[$key]['UNIT'] = $singleParam['UNIT'];
                            }
                        }
                    }
                    $device['CHANNELS'][$channel] = ["params" => $params, "values" => $this->hg->getParamset(intval($id),$channel,"MASTER")];
                } else {
                    unset($device['CHANNELS'][$channel]);
                }
            } catch(\Homegear\HomegearException $e) { }
        }
        return $device;
    }

    public function loadVariables($id = false)
    {
        if ($id === false) { return; }
        $device = $this->hg->getAllValues(intval($id))[0];
        foreach ($device['CHANNELS'] as $key => $channel) {
            // only get channels with params
            if (array_key_exists('PARAMSET', $channel) && is_array($channel['PARAMSET'])) {
                foreach ($channel['PARAMSET'] as $k => $singleParam) {
                    // check if param is visible & writeable
                    if ($singleParam['WRITEABLE'] !== true) {
                        unset($channel['PARAMSET'][$k]);
                    } else {
                        // fix encoding
                        if (array_key_exists('UNIT', $singleParam)) {
                            $channel['PARAMSET'][$k]['UNIT'] = $singleParam['UNIT'];
                        }
                    }
                }
                $device['CHANNELS'][$key]['PARAMSET'] = $channel['PARAMSET'];
            }
            if (count($device['CHANNELS'][$key]['PARAMSET']) == 0) {
                unset($device['CHANNELS'][$key]);
            }
        }
        return $device;
    }

    public function updateConfigParams($id = false, $params = false)
    {
        if ($id === false || $params === false) { return; }
        $channel = intval($params['channel']);
        $paramDesc = $this->hg->getParamsetDescription(intval($id),$channel,"MASTER");
        $values = $this->hg->getParamset(intval($id),$channel,"MASTER");
        foreach ($paramDesc as $param) {
            $paramId = $param['ID'];
            if (array_key_exists($paramId,$params)) {
                if ($param['TYPE'] == "BOOL") {
                    $values[$paramId] = boolval($params[$paramId]);
                } elseif ($param['TYPE'] == "ENUM") {
                    if (array_key_exists(intval($params[$paramId]),$param['VALUE_LIST'])) {
                        $values[$paramId] = intval($params[$paramId]);
                    }
                } elseif ($param['TYPE'] == "INTEGER") {
                    if (intval($params[$paramId]) >= $param['MIN'] && intval($params[$paramId]) <= $param['MAX']) {
                        $values[$paramId] = intval($params[$paramId]);
                    }
                } elseif ($param['TYPE'] == "FLOAT") {
                    if (floatval($params[$paramId]) >= $param['MIN'] && floatval($params[$paramId]) <= $param['MAX']) {
                        $values[$paramId] = floatval($params[$paramId]);
                    }
                } else {
                    $values[$paramId] = $params[$paramId];
                }
            }
        }
        $this->hg->putParamset(intval($id),$channel,$values);
    }

    public function updateValue($id = false, $params = false)
    {
        if ($id === false || $params === false) { return; }
        $channel = intval($params['channel']);
        if (array_key_exists('type',$params)) {
            // check type
            if ($params['type'] == 'BOOL') {
                $params['value'] = boolval($params['value']);
            } elseif ($params['type'] == 'FLOAT') {
                $params['value'] = floatval($params['value']);
            } elseif ($params['type'] == 'INTEGER' || $params['type'] == 'ENUM') {
                $params['value'] = intval($params['value']);
            } else {
                $params['value'] = strval($params['value']);
            }
        }
        try {
            $this->hg->setValue(intval($id),$channel,strval($params['variable']),$params['value']);
        } catch(\Homegear\HomegearException $e) { }
    }

    public function forceConfigUpdate($id = false)
    {
        if ($id === false) { return; }
        try {
            $this->hg->forceConfigUpdate(intval($id));
        } catch(\Homegear\HomegearException $e) {
            // do nothing
        }
        return 1;
    }

    public function loadHeatingProgram($id = false, $db = false, $global = false)
    {
        if ($id === false) { return; }
        if ($db === false) {
            // load program directly from device
            $data = $this->loadConfigParams($id);
            if (array_key_exists('FAMILY',$data) && $data['FAMILY'] == 666) {
                // We need to get the heating groups from hmip-cloud first
                $this->hg->setValue(intval($id), 0, "fetchHeatingProfiles");
                $data = $this->loadConfigParams($id);
            }
            if (in_array(strtoupper($data['TYPE']),$this->devicesHeatingProgram['single'])) {
                $data['HEATINGDATA'] = $this->_formatHeatingData($data,'single');
            } else {
                $data['HEATINGDATA_MULTIPLE'] = 1;
                $data['HEATINGDATA'] = $this->_formatHeatingData($data,'multiple');
            }
        }
        return $data;
    }

    public function loadHeatingProgramTemplates($id = false)
    {
        if ($id === false) { return ['private' => null, 'global' => null]; }
        $private = $this->hg->getData(self::componentHeatingProgramTemplates,$id);
        if (is_array($private)) {
            foreach ($private as $key => $value) {
                $tmp['CHANNELS'][0]['values'] = $private[$key]['data'];
                $private[$key]['data'] = $this->_formatHeatingData($tmp,'single');
            }
        }
        $global = $this->hg->getData(self::componentHeatingProgramTemplates,'global');
        if (is_array($global)) {
            foreach ($global as $key => $value) {
                $tmp['CHANNELS'][0]['values'] = $global[$key]['data'];
                $global[$key]['data'] = $this->_formatHeatingData($tmp,'single');
            }
        }
        return ['private' => $private, 'global' => $global];
    }

    public function saveHeatingProgramDevice($id = false, $data = false) 
    {
        if ($id === false || $data === false) { return 0; }
        $result = ['result' => 'error'];
        $chKey = 0;
        // get the channel with the needed settings
        $deviceData = $this->loadConfigParams($id);
        if (array_key_exists('FAMILY',$deviceData) && $deviceData['FAMILY'] == 666) {
            $data2save = array();
            $myData = $deviceData['CHANNELS'][0]['values']['heatingProfiles'];
            foreach ($myData as $profileName => $profileData) {
                $getName = $profileName;
                if (array_key_exists($profileName,$deviceData['CHANNELS'][0]['values']['profiles']) && $deviceData['CHANNELS'][0]['values']['profiles'][$profileName]['name'] != '') {
                    $getName = $deviceData['CHANNELS'][0]['values']['profiles'][$profileName]['name'];
                }
                $data2save[$profileName] = $myData[$profileName];
                foreach (['MONDAY', 'TUESDAY', 'WEDNESDAY', 'THURSDAY', 'FRIDAY', 'SATURDAY', 'SUNDAY'] as $day) {
                    unset($data2save[$profileName]['profileDays'][$day]['periods']);
                    $data2save[$profileName]['profileDays'][$day]['baseValue'] = $myData[$profileName]['profileDays'][$day]['baseValue'];
                    foreach ($data[$getName][$day] as $key => $value) {
                        if ($value[1] == $myData[$profileName]['profileDays'][$day]['baseValue']) { continue; }
                        $data2save[$profileName]['profileDays'][$day]['periods'][] = ['endtime' => date("H:i",date("U",strtotime($data[$getName][$day][$key+1][0]))), 'starttime' => date("H:i",date("U",strtotime($value[0]))), 'value' => floatval($value[1])];
                    }
                }
            }
            unset($deviceData['CHANNELS'][0]['values']['heatingProfiles']);
            $deviceData['CHANNELS'][0]['values']['heatingProfiles'] = $data2save;
            $this->hg->putParamset(intval($id), 0, ['heatingProfiles' => $deviceData['CHANNELS'][0]['values']['heatingProfiles']]);
            return ['result' => 'success'];
        }
        foreach ($deviceData['CHANNELS'] as $k => $v) {
            if ($v['params'] && count(preg_grep('/ENDTIME/',array_keys($v['params']))) > 0) {
                $chKey = $k;
                break;
            }
        }
        // check if we have multiple programs
        if (array_key_exists('P1',$data)) {
            foreach ($data as $key => $value) {
                $data2save = $this->_formatHeatingData2Save($data[$key],$key);
                if ($data2save != 0) {
                    try {
                        $this->hg->putParamset(intval($id), $chKey, $data2save);
                    } catch(\Homegear\HomegearException $e) {
                        // do nothing
                    }
                    $result = ['result' => 'success'];
                }
            }
        } else {
            $data2save = $this->_formatHeatingData2Save($data);
            if ($data2save != 0) {
                $this->hg->putParamset(intval($id), 0, $data2save);
                $result = ['result' => 'success'];
            }
        }
        return $result;
    }

    public function saveHeatingProgramTemplate($id = null, $data = null)
    {
        if ($id === null || $data === null) { return; }
        if (!array_key_exists('data', $data) || $data['data'] == '') { return; }
        // convert data
        $data2save = $this->_formatHeatingData2Save(json_decode($data['data'],true));
        if (array_key_exists('newTemplate', $data) && $data['newTemplate'] != '') {
            // new template
            if ($data['type'] == "private") {
                // template for the current device, get all templates first
                $templates = $this->hg->getData(self::componentHeatingProgramTemplates,$id);
                // add template
                if (is_array($templates)) {
                    array_push($templates, ['id' => $data['newTemplate'], 'name' => $data['newTemplate'], 'data' => $data2save]);
                } else {
                    $templates[] = ['id' => $data['newTemplate'], 'name' => $data['newTemplate'], 'data' => $data2save];
                }
                $this->hg->setData(self::componentHeatingProgramTemplates,$id,$templates);
            } else {
                // global template, get all templates first
                $templates = $this->hg->getData(self::componentHeatingProgramTemplates,'global');
                // add template
                if (is_array($templates)) {
                    array_push($templates, ['id' => $data['newTemplate'], 'name' => $data['newTemplate'], 'data' => $data2save]);
                } else {
                    $templates[] = ['id' => $data['newTemplate'], 'name' => $data['newTemplate'], 'data' => $data2save];
                }
                $this->hg->setData(self::componentHeatingProgramTemplates,'global',$templates);
            }
        } else if (array_key_exists('currentTemplate', $data) && $data['currentTemplate'] != '') {
            // update a template
            if ($data['type'] == "private") {
                // template for the current device, get all templates first
                $templates = $this->hg->getData(self::componentHeatingProgramTemplates,$id);
                // update template
                foreach ($templates as $key => $template) {
                    if ($template['id'] == $data['currentTemplate']) {
                        $templates[$key]['data'] = $data2save;
                    }
                }
                $this->hg->setData(self::componentHeatingProgramTemplates,$id,$templates);
            } else {
                // global template, get all templates first
                $templates = $this->hg->getData(self::componentHeatingProgramTemplates,'global');
                // update template
                foreach ($templates as $key => $template) {
                    if ($template['id'] == $data['currentTemplate']) {
                        $templates[$key]['data'] = $data2save;
                    }
                }
                $this->hg->setData(self::componentHeatingProgramTemplates,'global',$templates);
            }
        }
        return;
    }

    public function deleteHeatingProgramTemplate($id = null, $data = null)
    {
        if ($id === null || $data === null) { return; }
        if (!array_key_exists('templateId', $data) || $data['templateId'] == '') { return; }
        $myType = explode('_',$data['templateId']);
        $type = $myType[0];
        // check if we have a valid template-type
        if ($type == 'private') {
            $id2delete = str_replace('private_','',$data['templateId']);
            // get all templates for the device
            $templates = $this->hg->getData(self::componentHeatingProgramTemplates,$id);
            foreach ($templates as $key => $template) {
                if ($id2delete == $template['id']) {
                    unset($templates[$key]);
                }
            }
            $this->hg->setData(self::componentHeatingProgramTemplates,$id,$templates);
        } else {
            $id2delete = str_replace('global_','',$data['templateId']);
            // get all global templates
            $templates = $this->hg->getData(self::componentHeatingProgramTemplates,'global');
            foreach ($templates as $key => $template) {
                if ($id2delete == $template['id']) {
                    unset($templates[$key]);
                }
            }
            $this->hg->setData(self::componentHeatingProgramTemplates,'global',$templates);
        }
    }

    public function loadMetadata($id = null)
    {
        if ($id === null) { return; }
        $device = $this->hg->getDeviceDescription(intval($id),-1);
        $device['METADATA'] = $this->hg->getAllMetadata(intval($id));
        return $device;
    }

    public function setMetadata($id = null, $key = null, $value = '')
    {
        if ($id === null || $key === null) { return; }
        $this->hg->setMetadata(intval($id),"$key","$value");
    }

    public function deleteMetadata($id = null, $key = null)
    {
        if ($id === null || $key === null) { return; }
        $this->hg->deleteMetadata(intval($id),"$key");
    }

    private function _formatHeatingData($data = null, $type = null)
    {
        if ($data === null || $type === null) { return; }
        if (array_key_exists('FAMILY',$data) && $data['FAMILY'] == 666) {
            $result = array();
            $myData = $data['CHANNELS'][0]['values']['heatingProfiles'];
            foreach ($myData as $profileName => $profileData) {
                $newName = $profileName;
                if (array_key_exists($profileName,$data['CHANNELS'][0]['values']['profiles']) && $data['CHANNELS'][0]['values']['profiles'][$profileName]['name'] != '') {
                    $newName = $data['CHANNELS'][0]['values']['profiles'][$profileName]['name'];
                }
                foreach (['MONDAY', 'TUESDAY', 'WEDNESDAY', 'THURSDAY', 'FRIDAY', 'SATURDAY', 'SUNDAY'] as $day) {
                    $result[$newName][$day][] = ['time' =>  date('c', mktime(0,"0")), 'value' => $profileData['profileDays'][$day]['baseValue']];
                    foreach ($profileData['profileDays'][$day]['periods'] as $key => $value) {
                        $starttime = explode(':',$value['starttime']);
                        $endtime = explode(':',$value['endtime']);
                        $result[$newName][$day][] = ['time' =>  date('c', mktime($starttime[0],$starttime[1])), 'value' => $value['value']];
                        $result[$newName][$day][] = ['time' =>  date('c', mktime($endtime[0],$endtime[1])), 'value' => $profileData['profileDays'][$day]['baseValue']];
                    }
                    $result[$newName][$day][] = ['time' =>  date('c', mktime(0,1440)), 'value' => $profileData['profileDays'][$day]['baseValue']];
                }
            }
            return $result;
        }
        $chKey = null;
        // get the channel with the needed settings
        foreach ($data['CHANNELS'] as $k => $v) {
            if ($v['values'] && count(preg_grep('/ENDTIME/',array_keys($v['values']))) > 0) {
                $chKey = $k;
                break;
            }
        }
        if ($chKey === null) { return; }
        if ($type == 'single') {
            $count = array();
            $myArray = array();
            foreach ($data['CHANNELS'][$chKey]['values'] as $key => $value) {
                // get all endtimes first
                if (stristr($key,'ENDTIME')) {
                    $tmp = explode('_',$key);
                    $myArray[$tmp[1]][$tmp[2]] = $value;
                }
            }
            foreach($myArray as $weekDay => $myValue) {
                // sort entries
                ksort($myValue);
                foreach($myValue as $index => $value ) {
                    if (!array_key_exists($weekDay,$count)) {
                        $count[$weekDay] = 0;
                    }
                    if ($count[$weekDay] > 0) { continue; }
                    if ($value == '1440') {
                        // we only need one last endpoint
                        if (array_key_exists($weekDay,$count) && $count[$weekDay] > 0) {
                            continue;
                        }
                        $count[$weekDay] = 1;
                    }
                    $myKey = 'TEMPERATURE_'.$weekDay.'_'.(intval($index)+1);
                    if (!array_key_exists($myKey,$data['CHANNELS'][$chKey]['values'])) {
                        $myKey = 'TEMPERATURE_'.$weekDay.'_'.$index;
                    }
                    $result[$weekDay][$index] = ['time' => date('c', mktime(0,$value)), 'value' => $data['CHANNELS'][$chKey]['values'][$myKey]];
                    if (!array_key_exists(0,$result[$weekDay])) {
                        $result[$weekDay][0] = ['time' => date('c', mktime(0,"0")), 'value' => $data['CHANNELS'][$chKey]['values']['TEMPERATURE_'.$tmp[1].'_1']];
                    }
                }
                if (count($result[$weekDay]) == 2) {
                    $result[$weekDay][1]['value'] = $result[$weekDay][0]['value'];
                }
            }
        }
        if ($type == 'multiple') {
            $count = array();
            $myArray = array();
            foreach ($data['CHANNELS'][$chKey]['values'] as $key => $value) {
                // get all endtimes first
                if (stristr($key,'ENDTIME')) {
                    $tmp = explode('_',$key);
                    $myArray[$tmp[0]][$tmp[2]][$tmp[3]] = $value;
                }
            }
            foreach($myArray as $programm => $P) {
                foreach($P as $weekDay => $myValue) {
                    // sort entries
                    ksort($myValue);
                    foreach($myValue as $index => $value ) {
                        if (!array_key_exists($programm,$count)) { $count[$programm] = []; }
                        if (!array_key_exists($weekDay,$count[$programm])) {
                            $count[$programm][$weekDay] = 0;
                        }
                        if ($count[$programm][$weekDay] > 0) { continue; }
                        if ($value == '1440') {
                            // we only need one last endpoint
                            if (array_key_exists($weekDay,$count[$programm]) && $count[$programm][$weekDay] > 0) {
                                continue;
                            }
                            $count[$programm][$weekDay] = 1;
                        }
                        $myKey = $programm.'_TEMPERATURE_'.$weekDay.'_'.(intval($index)+1);
                        if (!array_key_exists($myKey,$data['CHANNELS'][$chKey]['values'])) {
                            $myKey = $programm.'_TEMPERATURE_'.$weekDay.'_'.$index;
                        }
                        $result[$programm][$weekDay][$index] = ['time' => date('c', mktime(0,$value)), 'value' => $data['CHANNELS'][$chKey]['values'][$myKey]];
                        if (!array_key_exists(0,$result[$programm][$weekDay])) {
                            $result[$programm][$weekDay][0] = ['time' => date('c', mktime(0,"0")), 'value' => $data['CHANNELS'][$chKey]['values'][$programm.'_TEMPERATURE_'.$weekDay.'_1']];
                        }
        
                    }
                }
            }
        }
        if ($type == 'single') {
            foreach ($result as $key => $value) {
                if (count($value) == 2) {
                    $result[$key][1]['value'] = $result[$key][0]['value'];
                }
                ksort($result[$key]);
            }
            foreach ($this->weekdays as $day) {
                if (array_key_exists(strtoupper($day),$result)) {
                    $tmp1[strtoupper($day)] = $result[strtoupper($day)];
                }
            }
        }
        if ($type == 'multiple') {
            foreach ($result as $key => $value) {
                foreach ($value as $k => $v) {
                    if (count($v) == 2) {
                        $result[$key][$k][1]['value'] = $result[$key][$k][0]['value'];
                    }
                    ksort($result[$key][$k]);
                }
            }
            foreach ($this->weekdays as $day) {
                foreach ($result as $key => $value) {
                    if (array_key_exists(strtoupper($day),$value)) {
                        $tmp1[$key][strtoupper($day)] = $result[$key][strtoupper($day)];
                    }
                }
            }
        }
        $result = $tmp1;
        return $result;
    }

    private function _formatHeatingData2Save($data = null, $program = null)
    {
        if ($data === null) { return 0; }
        foreach ($data as $key => $value) {
            foreach($value as $k => $v) {
                if (array_key_exists(($k+1),$value)) {
                    $myHours = date("G",strtotime($value[$k+1][0]));
                    $myMinutes = date("i",strtotime($value[$k+1][0]));
                    if ($myHours == 0) {
                        $myTime = 1440;
                    } else {
                        $myTime = ($myHours * 60) + $myMinutes;
                    }
                } else {
                    $myTime = 1440;
                }
                if ($program === null) {
                    $data2save['TEMPERATURE_'.$key.'_'.($k+1)] = floatval($v[1]);
                    $data2save['ENDTIME_'.$key.'_'.($k+1)] = $myTime;
                } else {
                    $data2save[$program.'_TEMPERATURE_'.$key.'_'.($k+1)] = floatval($v[1]);
                    $data2save[$program.'_ENDTIME_'.$key.'_'.($k+1)] = $myTime;
                }
            }
            if (count($value) < 13) {
                // 13 values are hardcoded on the HM devices
                if ($program === null) {
                    $last = $data2save['ENDTIME_'.$key.'_'.count($value)];
                    for ($i = count($value); $i <= 13; $i++) {
                        $data2save['TEMPERATURE_'.$key.'_'.$i] = $last;
                        $data2save['ENDTIME_'.$key.'_'.$i] = 1440;
                    }
                } else {
                    $last = $data2save[$program.'_ENDTIME_'.$key.'_'.count($value)];
                    for ($i = count($value); $i <= 13; $i++) {
                        $data2save[$program.'_TEMPERATURE_'.$key.'_'.$i] = $last;
                        $data2save[$program.'_ENDTIME_'.$key.'_'.$i] = 1440;
                    }
                }
            }
        }
        return $data2save;
    }

    /**
     * Update a device
     * 
     * @return bool
     */
    public function update($id, $config) {
        // get full paramset
        try {
            $paramset = $this->hg->getParamset(intval($id),0);
        } catch(\Homegear\HomegearException $e) {  
            return $e;
        }
        // configure/update the device
        foreach ($config as $k => $v) {
            if ($k == 'DEVICE_NAME') {
                // set the device name
                try {
                    $this->hg->setName(intval($id),$v);
                } catch(\Homegear\HomegearException $e) {
                    return $e;
                }
            } else if (($k == 'PASSWORD') && ($v == '')) {
                // don't try to set empty password
                continue;
            } else if ($k == 'CENTRAL_ADDRESS_SPOOFED') {
                try {
                    $this->hg->setValue(intval($id),0,"CENTRAL_ADDRESS_SPOOFED",($v == 1 ? true : false));
                } catch(\Homegear\HomegearException $e) {
                    return $e;
                }
            } else if ($k == 'GATEWAY') {
                $family = $this->hg->getDeviceDescription(intval($id),-1,array("FAMILY"))['FAMILY'];
                if (in_array($family,self::mayChangeGateway)) {
                    try {
                        $this->hg->setInterface(intval($id),"$v");
                    } catch(\Homegear\HomegearException $e) {
                        return $e;
                    }
                }
            } else {
                try {
                    if (array_key_exists(strtoupper($k),$paramset)) {
                        // check type
                        if (is_bool($paramset[strtoupper($k)])) { $v = boolval($v); }
                        if (is_float($paramset[strtoupper($k)])) { $v = floatval($v); }
                        if (is_int($paramset[strtoupper($k)])) { $v = intval($v); }
                        if (is_array($paramset[strtoupper($k)])) { $v = print_r($v,true); }
                    }
                    $this->hg->putParamset(intval($id),0,"MASTER",array(strtoupper($k) => $v));
                } catch(\Homegear\HomegearException $e) {
                    return $e;
                }
            }
        }
        return 1;
    }

    public function updateChannelVisibility($id = null, $params = null)
    {
        if ($id === null || $params === null || !is_array($params)) { return; }
        $this->hg->setMetadata(intval($id), "showChannelNames", $params);
        return 1;
    }

    public function updateVariableVisibility($id = null, $params = null)
    {
        if ($id === null || $params === null || !is_array($params)) { return; }
        $this->hg->setMetadata(intval($id), "showVariableNames", $params);
        return 1;
    }

    /**
     * Set a name of a specific channel of a device
     * 
     * @return none
     */
    public function setChannelName($id = null, $channel = null, $name = null)
    {
        if ($id === null || $channel === null || $name === null) { return; }
        $this->hg->setName(intval($id), intval($channel), $name);
        return true;
    }

    /**
     * Delete a specific device
     * 
     * @return bool
     */
    public function delete($id)
    {
        try {
            // 5 = don't delete device from underlying gateway
            // 13 = delete device from underlying gateway
            //$this->hg->deleteDevice(intval($id),-1);
            $this->hg->deleteDevice(intval($id),5);
        } catch(\Homegear\HomegearException $e) {
            return $e;
        }
        return true;
    }

    /**
     * Delete all devices associated with a specific gateway
     * 
     * @return bool
     */
    public function deleteFromGateway($familyId = null, $gateway = null, $gateways = null)
    {
        if ($familyId === null || $gateway === null || $gateways === null) { return false; }
        $isOnly = true;
        // check if the gateway is the only one for this family
        if (is_array($gateways) && array_key_exists($familyId,$gateways)) {
            if (is_array($gateways[$familyId]) && count($gateways[$familyId]) > 1) {
                $isOnly = false;
            }
        }
        // get all devices
        $devices = $this->loadAll();
        foreach ($devices as $device) {
            // check if device matches desired family
            if (is_array($device) && array_key_exists('familyId',$device) && $device['familyId'] == $familyId) {
                // check if there is only one gateway left, needed for kxn
                if ($familyId == 14 && $isOnly === true) {
                    $this->delete($device['ID']);
                } else {
                    // check if device is connected via desired gateway
                    if (array_key_exists('gateway',$device) && $device['gateway'] == $gateway) {
                        $this->delete($device['ID']);
                    }
                }
            }
        }
        return true;
    }

    /**
     * List all known device families (used for adding devices)
     * 
     * @return array()
     */
    public function listAllKnownFamilies() {
        // return all loaded families
        return $this->hg->listFamilies();
    }

    /**
     * Check if a gateway is needed for adding a device. If yes, return the known gateway(s)
     * 
     * @return array()
     */
    public function getGatewaysForFamily($familyId = null)
    {
        if ($familyId === null) { return; }
        // check if the selected family needs a gateway
        $info = $this->hg->getPairingInfo(intval($familyId));
        if (array_key_exists('interfaces',$info)) {
            // check if this is only an empty array
            if (count($info['interfaces']) == 0) {
                return ['result' => '0'];
            }
            // get all available gateway for this family
            $gateways = $this->hg->listInterfaces(intval($familyId));
            foreach ($gateways as $key => $gateway) {
                $alias = $this->hg->getData(self::componentGatewayAlias,$familyId.'-'.$gateway['ID']);
                $gateways[$key]['GATEWAY_ALIAS'] = (isset($alias) ? $alias : '');
            }
            return ['result' => '1', 'gateways' => $gateways];
        } else {
            return ['result' => '0'];
        }
    }

    /**
     * Get all devices of a specific family for adding.
     * 
     * @return array()
     */
    public function getDevicesForFamily($familyId = null)
    {
        if ($familyId === null) { return; }
        $info = $this->hg->getPairingInfo(intval($familyId));
        if (array_key_exists('pairingMethods',$info)) {
            // new variant
            if (!array_key_exists('createDevice',$info['pairingMethods'])) {
                // we don't have to create the device manually
                $items = $this->checkForTypeSelector($info);
                if (!$items) {
                    $familyNames = $this->hg->listFamilies();
                    $key = array_search($familyId, array_column($familyNames, 'ID'));
                    $items[] = ['TYPE_ID' => '-1', 'TYPE' => $familyNames[$key]['NAME']." - Device"];
                }
            } else {
                $items = $this->hg->listKnownDeviceTypes(intval($familyId),false,array('TYPE','TYPE_ID','NAME','DESCRIPTION'));
            }
        } else {
            // depricated variant
            if (!in_array('createDevice',$info)) {
                // we don't have to create the device manually
                $familyNames = $this->hg->listFamilies();
                $key = array_search($familyId, array_column($familyNames, 'ID'));
                $items[] = ['TYPE_ID' => '-1', 'TYPE' => $familyNames[$key]['NAME']." - Device"];
            } else {
                $items = $this->hg->listKnownDeviceTypes(intval($familyId),false,array('TYPE','TYPE_ID','NAME','DESCRIPTION'));
            }
        }
        usort($items, function($a, $b) { return $a['TYPE'] <=> $b['TYPE']; });
        return $items;
    }

    /**
     * Get information for adding a specific device
     * 
     * @return array()
     */
    public function getDeviceDataForAdding($family = null, $type = null, $gateway = null)
    {
        if ($family === null || $type === null) { return; }
        $device['FAMILY'] = $family;
        $device['TYPE_ID'] = $type;
        $device['GATEWAY'] = $gateway;
        // Exception: EnOcean and Intertechno needs sniffing!
        if ($family == 15 || $family == 16 || $family == 420 || $family == 427) { $device['SNIFFING'] = 1; }
        // get pairing methods
        $pairingMethods = $this->hg->getPairingInfo(intval($family));
        // check pairingMethods if module supports sniffing
        if ($family == 420) {
            if (!array_key_exists('sniffDevices',$pairingMethods['pairingMethods'])) {
                $device['SNIFFING'] = 0;
            }
        }
        if ($family == 427) {
            // for RTSP, sniffing is only available when using ONVIF
            if ($type == 1) { $device['SNIFFING'] = 0; }
        }
        // check if we're adding a specific device
        if ($type != "-1" && is_numeric($type)) {
            // check if we have device-specific pairing methods (eg EnOcean)
            // get all known devices for this family (needed to filter for desired pairing method)
            $allDevices = $this->hg->listKnownDeviceTypes(intval($family),false,array('TYPE_ID','PAIRING_METHOD'));
            // search & get the current device
            $myDevice = $allDevices[array_search($type, array_column($allDevices, 'TYPE_ID'))];
            if (isset($myDevice) && $myDevice !== null && array_key_exists('PAIRING_METHOD',$myDevice) && $myDevice['PAIRING_METHOD'] != '') {
                // we have our device + pairing method!
                if (is_array($myDevice['PAIRING_METHOD'])) {
                    foreach ($myDevice['PAIRING_METHOD'] as $key => $method) {
                        // get pairingMethod from XML, may be extended with typeSelector for getPairingInfo of family!
                        $myTypeSelector = false;
                        $myPairingMethod = explode('+',$method);
                        $myDevice['PAIRING_METHOD'] = $myPairingMethod[0];
                        if (array_key_exists(1,$myPairingMethod)) {
                            $myTypeSelector = preg_replace('/[\[\]]/','',$myPairingMethod[1]);
                        }
                        // Only support sniffing with 'createDevice'
                        if ($myPairingMethod[0] != 'createDevice') { $device['SNIFFING'] = 0; } 
                        // "normalize" createDevice+
                        if (strpos($method, 'createDevice+') !== false) { $myPairingMethod[0] = 'createDevice+'; }
                        $device['PAIRING_METHODS'][$myPairingMethod[0]]['CONFIG_PARARMS'] = $this->_getPairingMethodData($family,$type,$myPairingMethod[0],$myTypeSelector,$pairingMethods['pairingMethods']);
                    }
                } else {
                    // get pairingMethod from XML, may be extended with typeSelector for getPairingInfo of family!
                    $myTypeSelector = false;
                    $myPairingMethod = explode('+',$myDevice['PAIRING_METHOD']);
                    $myDevice['PAIRING_METHOD'] = $myPairingMethod[0];
                    if (array_key_exists(1,$myPairingMethod)) {
                        $myTypeSelector = preg_replace('/[\[\]]/','',$myPairingMethod[1]);
                    }
                    // Only support sniffing with 'createDevice'
                    if ($myDevice['PAIRING_METHOD'] != 'createDevice') { $device['SNIFFING'] = 0; } 
                    // "normalize" createDevice+
                    if (strpos($myDevice['PAIRING_METHOD'], 'createDevice+') !== false) { $myDevice['PAIRING_METHOD'] = 'createDevice+'; }
                    $device['PAIRING_METHODS'][$myDevice['PAIRING_METHOD']]['CONFIG_PARARMS'] = $this->_getPairingMethodData($family,$type,$myDevice['PAIRING_METHOD'],$myTypeSelector,$pairingMethods['pairingMethods']);
                }
                return $device;
            }
        }
        // we don't have device-specific paring methods, or do not add a specific device
        if (array_key_exists('pairingMethods',$pairingMethods)) {
            foreach ($pairingMethods['pairingMethods'] as $method => $value) {
                if (is_array($value)) {
                    if ($method == "searchDevices") {
                        // search devices
                        if ($family == 24) { continue; }
                        $device['PAIRING_METHODS']['searchDevices'] = ['name' => 'searchDevices'];
                    }
                    if ($method == "setInstallMode") {
                        // set install mode
                        $device['PAIRING_METHODS'][$method]['CONFIG_PARARMS'] = $this->_getPairingMethodData($family,$type,$method,null,$pairingMethods['pairingMethods']);
                    }
                    if ($method == "addDevice") {
                        // add device
                        $device['PAIRING_METHODS']['addDevice'] = ['name' => 'addDevice'];
                        $device['PAIRING_METHODS']['addDevice']['CONFIG_PARARMS'][] = ['id' => 'serialNumber', 'label' => 'Serial number', 'decsription' => 'The serial number of the device to be paired'];
                    }
                    if ($method == "createDevice") {
                        // create device
                        if ($type == -1) { continue; }
                        $device['PAIRING_METHODS'][$method]['CONFIG_PARARMS'] = $this->_getPairingMethodData($family,$type,$method,null,$pairingMethods['pairingMethods']);
                    }
                    if ($method == "createDevice+") {
                        // create device +
                        if ($type == -1) { continue; }
                        $device['PAIRING_METHODS'][$method]['CONFIG_PARARMS'] = $this->_getPairingMethodData($family,$type,$method,null,$pairingMethods['pairingMethods']);
                    }
                }
            }
            
        }
        return $device;
    }

    private function _getPairingMethodData($family = null, $type = null, $method = null, $myTypeSelector = null, $pairingMethods = array())
    {
        if ($family === null || $type === null || $method === null || count($pairingMethods) < 1) { return; }
        $result = array();
        $config = $this->getDeviceConfigParams($family,$type);
        if (array_key_exists('CONFIG_PARARMS',$config) && is_array($config['CONFIG_PARARMS'])) {
            foreach ($config['CONFIG_PARARMS'] as $key => $value) {
                if (array_key_exists('label', $value) && $value['label'] == '') {
                    $value['label'] = $family.'_label_'.strtolower($value['id']);
                }
                if (array_key_exists('description', $value) && $value['description'] == '') {
                    $value['description'] = $family.'_desc_'.strtolower($value['id']);
                }
                $result[$key] = $value;
            }
        }
        if (array_key_exists($method, $pairingMethods)) {
            // device paring method matches one of the family pairing methods
            // filter allready selected fields
            if ($myTypeSelector !== null && $myTypeSelector == '-eep') {
                $myType = preg_replace('/-/','',$config['TYPE']);
                $result[] = array('id' => 'eep', 'value' => $myType, 'type' => 'hidden');
                return $result;
            }
            if (array_key_exists('typeSelector', $pairingMethods[$method]['metadataInfo'])) {
                foreach ($pairingMethods[$method]['metadataInfo']['typeSelector'] as $key => $data) {
                    if ($myTypeSelector !== null && $key == $myTypeSelector) {
                        if (array_key_exists('additionalFields', $data) && is_array($data['additionalFields']) && count($data['additionalFields']) > 0) {
                            foreach($data['additionalFields'] as $fieldkey => $field) {
                                $result[$field['pos']] = array('id' => $fieldkey, 'label' => $field['name'], 'description' => $field['description'], 'type' => $field['type']);
                            }
                            $result[] = array('id' => 'type', 'value' => $myTypeSelector, 'type' => 'hidden');
                            ksort($result);
                            return $result;
                        }
                    }
                }
            } else if (array_key_exists('fields',$pairingMethods[$method]['metadataInfo'])) {
                foreach ($pairingMethods[$method]['metadataInfo']['fields'] as $fkey => $field) {
                    // we allready know the deviceType & interfaceId
                    if ($field == "deviceType") { unset($pairingMethods[$method]['metadataInfo']['fields'][$fkey]); continue; }
                    if ($field == "interfaceId") { unset($pairingMethods[$method]['metadataInfo']['fields'][$fkey]); continue; }
                    if (is_array($pairingMethods[$method]['metadataInfo']['fields'])) {
                        $result[] = ['id' => $field, 'label' => $family.'_label_'.$field, 'type' => '', 'description' => $family.'_desc_'.$field ];
                    } else {
                        $result[$fkey] = ['id' => $field, 'label' => $family.'_label_'.$field, 'type' => '', 'description' => $family.'_desc_'.$field ];
                    }
                }
                ksort($result);
                return $result;
            }
        }
    }

    /**
     * Start sniffing for device packages
     * 
     * @return none
     */
    public function startSniffing($family = null)
    {
        if ($family == null) { return; }
        $this->hg->startSniffing(intval($family));
    }

    /**
     * Stop sniffing for device packages and return packages
     * 
     * @return array()
     */
    public function stopSniffing($family = null)
    {
        if ($family == null) { return; }
        $result = $this->hg->getSniffedDevices(intval($family));
        // this is hardcoded! atm only enocean supports sniffing, and the addresses are needed as hex
        foreach ($result as $key => $value) {
            if ($family == 15) {
                $result[$key]['ADDRESS'] = strtoupper(dechex((float)$value['ADDRESS']));
            }
        }
        $this->hg->stopSniffing(intval($family));
        return $result;
    }

    /**
     * Get config parameters
     * 
     * @return array()
     */
    public function getDeviceConfigParams($family = null, $type = null, $gateway = null)
    {
        // Todo: need refactoring!
        if (array_key_exists('locale',$_SESSION) && is_array($_SESSION['locale'])) {
            $this->hg->setLanguage($_SESSION['locale'][0]);
        }
        if ($family == null || $type == null) { return; }
        $knownDevices = $this->hg->listKnownDeviceTypes(false,array('FAMILY','TYPE','TYPE_ID','DESCRIPTION','SERIAL_PREFIX','CHANNELS'));
        $key = $this->multidimensional_search($knownDevices, array('FAMILY'=>$family, 'TYPE_ID'=>$type));
        $knownDevices[$key]['PAIRING_METHODS'] = $this->hg->getPairingInfo(intval($family))['pairingMethods'];
        // check if the devices need to be created manually
        if (array_key_exists('CHANNELS', $knownDevices[$key]) && (array_key_exists('createDevice', $knownDevices[$key]['PAIRING_METHODS']) || array_key_exists('createDevice+', $knownDevices[$key]['PAIRING_METHODS']))) {
            foreach ($knownDevices[$key]['CHANNELS'] as $channel) {
                try {
                    // get paramsets for device, needed for generating the html form within "add device"
                    $paramset = $this->hg->getParamsetDescription(intval($knownDevices[$key]['FAMILY']), intval($knownDevices[$key]['TYPE_ID']),-1,$channel,"MASTER");
                    foreach ($paramset as $param) {
                        // check if channel has form fields
                        //if (array_key_exists('LABEL',$param) && array_key_exists('FORM_FIELD_TYPE',$param) && array_key_exists('FORM_POSITION',$param)) {
                        if (array_key_exists('FORM_FIELD_TYPE',$param) && array_key_exists('FORM_POSITION',$param)) {
                            if ($param['FORM_FIELD_TYPE'] == 'select') {
                                $param['LABEL'] = (isset($param['LABEL']) ? $param['LABEL'] : $family.'_label_'.strtolower($param['ID']));
                                $knownDevices[$key]['CONFIG_PARARMS'][$param['FORM_POSITION']] = array('id' => $param['ID'],'label' => $param['LABEL'], 'description' => (isset($param['DESCRIPTION']) ? $param['DESCRIPTION'] : $family.'_desc_'.strtolower($param['ID'])), 'type' => $param['FORM_FIELD_TYPE'], 'values' => $param['VALUE_LIST'], 'position' => $param['FORM_POSITION']);
                            } else {
                                $param['LABEL'] = (isset($param['LABEL']) ? $param['LABEL'] : $family.'_label_'.strtolower($param['ID']));
                                $knownDevices[$key]['CONFIG_PARARMS'][$param['FORM_POSITION']] = array('id' => $param['ID'],'label' => $param['LABEL'], 'description' => (isset($param['DESCRIPTION']) ? $param['DESCRIPTION'] : $family.'_desc_'.strtolower($param['ID'])), 'type' => $param['FORM_FIELD_TYPE'], 'position' => $param['FORM_POSITION']);
                            }
                        }
                    }
                } catch(\Homegear\HomegearException $e) {
                    #return ['result' => 'error', 'message' => $e->getMessage()];
                }
            }
        }
        return $knownDevices[$key];
    }

    /**
     * Pair a device (needed for EASYLED)
     * 
     * @return none
     */
    public function pair($id = null, $channel = null) {
        // pair device to device
        if ($id === null || $channel === null) { return; }
        $this->hg->setValue((int)$id, (int)$channel, 'PAIRING', true, false);
    }
    
    /**
     * Unpair a device (needed for EASYLED)
     * 
     * @return none
     */
    public function unpair($id = null, $channel = null) {
        // unpair device from device
        if ($id === null || $channel === null) { return; }
        $this->hg->setValue((int)$id, (int)$channel, 'UNPAIRING', true, false);
    }

    /**
     * Search for new devices
     * 
     * @return int
     */
    public function searchDevices($familyId = null)
    {
        if ($familyId === null) { return; }
        return $this->hg->searchDevices(intval($familyId));
    }

    /**
     * Create a new device
     * 
     * @return int
     */
    public function createDevice($familyId = null, $typeId = null, $config = array())
    {
        if ($familyId === null || $typeId === null) { return; }
        $snr = '';
        if (array_key_exists('address',$config) && $config['address'] == '') { return; }
        // create a serial for the new device if there is none
        if ((array_key_exists('SERIAL',$config) && $config['SERIAL'] == '') || (array_key_exists('serialNumber',$config) && $config['serialNumber'] == '')) {
            $snr = $this->_create_serial('ESY');
        } else {
            if (array_key_exists('SERIAL',$config)) {
                $snr = $config['SERIAL'];
            }
            if (array_key_exists('serialNumber',$config)) {
                $snr = $config['serialNumber'];
            }
        }
        // try to create the new device
        try {
            $address = (array_key_exists('address',$config) ? $config['address'] : -1);
            // this is enocean, the address is hex ... so decode it first
            if ($familyId == 15) { $address = hexdec($address); }
            if (!array_key_exists('gateway',$config)) {
                // temp. hack 'cause needed infos are missing from homegear
                // EASYLed/2 must provide an ip-address, convert to long()
                if ($familyId == 8 || $familyId == 18) { $address = ip2long($address); }
                $newID = $this->hg->createDevice(intval($familyId),intval($typeId),"$snr",intval($address),-1);
            } else {
                $newID = $this->hg->createDevice(intval($familyId),intval($typeId),"$snr",intval($address),-1,$config['gateway']);
            }
        } catch(\Homegear\HomegearException $e) {
            return ['result' => 'error', 'message' => $e->getMessage()];
        }
        // configure the device
        if ($this->update($newID,$config) != 1) {
            // something went totally wrong. delete the whole device!
            $this->delete($newID);
            return ['result' => 'error'];
        }
        return ['result' => 'success', 'message' => $this->translator->trans("devices.created_success_id",array('%newId%' => $newID,'%edit%' => $this->router->generate('admin.inventory.devices.edit', ['id' => $newID])),'admin')];
    }

    public function setInstallMode($familyId = null, $typeId = null, $gateway = null, $params = null)
    {
        if ($familyId === null || $typeId === null || $gateway === null) { return; }
        if ($familyId == 24) {
            // setInstallMode for CCU2
            try {
                $this->hg->setInstallMode(intval($familyId), true, 60, array("interface" => "$gateway", "type" => "$typeId"));
            } catch(\Homegear\HomegearException $e) {
                return;
            }
        } else {
            $installParams = array();
            if ($gateway !== null) {
                $installParams['interface'] = "$gateway";
            }
            if ($params !== null) {
                if (array_key_exists('type',$params)) {
                    $installParams['type'] = $params['type'];
                }
                if (array_key_exists('eep',$params)) {
                    $installParams['eep'] = hexdec($params['eep']);
                }
                if (array_key_exists('config', $params)) {
                    if (is_array($params['config'])) {
                        foreach ($params['config'] as $key => $value) {
                            $installParams[$key] = $value;
                        }
                    }
                }
            }
            try {
                if (is_array($installParams) && count($installParams) >0) {
                    $this->hg->setInstallMode(intval($familyId), true, 60, $installParams);
                } else {
                    $this->hg->setInstallMode(intval($familyId), true, 60);
                }
            } catch(\Homegear\HomegearException $e) {
                return;
            }
        }
    }


    public function getPairingState($familyId = null)
    {
        if ($familyId === null) { return ['status' => 'error', 'message' => $this->translator->trans("devices.family_id_missing",array(),'admin')]; }
        $result = ['status' => 'error', 'message' => $this->translator->trans("devices.family_not_found",array(),'admin')];
        $modules = $this->hg->listModules();
        foreach ($modules as $module) {
            if ($module['FAMILY_ID'] == intval($familyId)) {
                if ($module['LOADED'] !== false) {
                    $result = $this->hg->getPairingState(intval($familyId));
                    if (array_key_exists('pairingModeEnabled', $result) && $result['pairingModeEnabled'] === true) {
                        $result['pairingModeEndInSeconds'] = $result['pairingModeEndTime'] - time();
                        break;
                    } else {
                        foreach ($result as $key => $value) {
                            if (is_int($key) && array_key_exists('state',$result[$key]) && $result[$key]['state'] == 'success') {
                                // we have a new device!
                                return ['status' => 'success', 'message' => $this->translator->trans("devices.found_success_id",array('%newId%' => $key,'%edit%' => $this->router->generate('admin.inventory.devices.edit', ['id' => $key])),'admin')];
                            }
                        }
                        return ['status' => 'success', 'message' => '0'];
                    }
                } else {
                    $result = ['status' => 'error', 'message' => $this->translator->trans("devices.family_not_loaded",array(),'admin')];
                    break;
                }
            }
        }
        return $result;
    }

    public function getMayChangeGateway()
    {
        return self::mayChangeGateway;
    }

    private function checkForTypeSelector($data = null)
    {
        if ($data === null || !is_array($data)) { return; }
        $items = array();
        foreach($data['pairingMethods'] as $key => $value) {
            if (is_array($value) && count($value) > 0 && array_key_exists('metadataInfo', $value) && array_key_exists('typeSelector',$value['metadataInfo'])) {
                foreach($value['metadataInfo']['typeSelector'] as $k => $v) {
                    $items[] = ['TYPE_ID' => $k, 'TYPE' => $v['name']];
                }
            }
        }
        return $items;
    }

    public function getDeviceFrontendHidden($deviceId = null)
    {
        if ($deviceId === null) { return; }
        return $this->hg->getData(self::componentDeviceFrontendHidden,"$deviceId");
    }

    public function setDeviceFrontendHidden($deviceId = null, $status = false)
    {
        if ($deviceId === null) { return; }
        if ($status === false) {
            $this->hg->deleteData(self::componentDeviceFrontendHidden,"$deviceId");
        } else {
            $this->hg->setData(self::componentDeviceFrontendHidden,"$deviceId",true);
        }
    }

    private function _create_serial($prefix) {
        return substr($prefix.random_int(10000000, 99999999),0,10);
    }

}